import inspect
from contextlib import suppress
from functools import wraps
from validateit.utils import check_type


def typeforced(callable):
    spec = inspect.getfullargspec(callable)

    def check_types(*args, **kwargs):
        fields = dict(zip(spec.args, args))
        fields.update(kwargs)
        for name, value in fields.items():
            with suppress(KeyError):
                type_hint = spec.annotations[name]
                if not check_type(type_hint, value):
                    raise TypeError(
                        "Unexpected type for '{}' (expected {} but found {})".format(
                            name, type_hint, type(value)
                        )
                    )

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            check_types(*args, **kwargs)
            return func(*args, **kwargs)

        return wrapper

    if inspect.isclass(callable):
        callable.__init__ = decorate(callable.__init__)
        return callable

    return decorate(callable)
